#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['gidgethub']

package_data = \
{'': ['*']}

install_requires = \
['uritemplate>=3.0.1', 'PyJWT[crypto]']

extras_require = \
{'aiohttp': ['aiohttp'],
 'dev': ['aiohttp',
         'black',
         'coverage[toml]>=5.0.3',
         'httpx',
         'mypy',
         'pytest-cov',
         'pytest-xdist',
         'tornado'],
 'doc': ['sphinx'],
 'httpx': ['httpx>=0.11.0'],
 'test': ['importlib-resources',
          'pytest>=5.4.1',
          'pytest-asyncio',
          'pytest-tornasync'],
 'tornado': ['tornado']}

setup(name='gidgethub',
      version='4.2.0',
      description='An async GitHub API library',
      author='Brett Cannon',
      author_email='brett@python.org',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
